﻿using System;
using System.Diagnostics;
using System.Collections.Generic;

namespace BeeDevelopment.Infrared.Protocols {

	/// <summary>
	/// Decodes incoming JVC commands.
	/// </summary>
	class JvcProtocolHandler : IRemoteControlProtocolHandler {

		#region Constants

		/// <summary>
		/// The minimum start bit length (nominally 8.4mS).
		/// </summary>
		const double MinimumStartBurstLength = 7.9E-3d;
		
		/// <summary>
		/// The maximum start bit length (nominally 8.4mS).
		/// </summary>
		const double MaximumStartBurstLength = 8.9E-3d;

		/// <summary>
		/// The threshold length between a short bit and a long bit.
		/// </summary>
		const double DataBitLengthThreshold = (1.05E-3d + 2.10E-3d) / 2.0d;

		/// <summary>
		/// The maximum time length between repeating commands (nominally 110mS - repeat pulse).
		/// </summary>
		const double RepeatCommandMaxLength = 130E-3;

		/// <summary>
		/// The maximum time length between data bits.
		/// </summary>
		private const double IntraBitMaxLength = 3.0E-3;

		#endregion

		#region Private Fields

		private double BurstLength = 0;

		private bool ReceivingStartBit = false;

		/// <summary>
		/// Stores the incoming decoded bits.
		/// </summary>
		private JvcCommand Command = null;

		/// <summary>
		/// Stores the last received command.
		/// </summary>
		private JvcCommand LastCommand = null;

		/// <summary>
		/// Stores the number of times the received command has been repeated.
		/// </summary>
		private int LastCommandRepeatCount = 0;

		/// <summary>
		/// A <see cref="Stopwatch"/> instance used to time repeating commands.
		/// </summary>
		private Stopwatch RepeatTimer = null;
		
		/// <summary>
		/// Allows the protocol handler to talk to its parent receiver.
		/// </summary>
		private RemoteControlReceiver Receiver;

		#endregion

		#region Constructor

		/// <summary>
		/// Create an instance of the <see cref="JvcProtocolHandler"/>.
		/// </summary>
		public JvcProtocolHandler(RemoteControlReceiver receiver) {
			this.Receiver = receiver;
			this.RepeatTimer = new Stopwatch();
			this.RepeatTimer.Start();
		}

		#endregion

		#region Protocol handler

		/// <summary>
		/// Handle the change in input pin signal state.
		/// </summary>
		/// <param name="time">The time, in seconds, since the pin last changed state.</param>
		/// <param name="newState">The new state of the pin.</param>
		public void PinChanged(double time, bool newState) {
			if (newState) {
				// We've just received a burst.
				this.BurstLength = time;
				if (time > JvcProtocolHandler.MinimumStartBurstLength && time < JvcProtocolHandler.MaximumStartBurstLength) {
					this.Command = null;
					this.LastCommand = null;
					this.ReceivingStartBit = true;
				} else {
					this.ReceivingStartBit = false;
					// Have we received everything?
					if (this.Command != null && this.Command.Count == 16) {
						if (!this.Receiver.WaitForSignalLevel(false, JvcProtocolHandler.IntraBitMaxLength)) {
							if (this.LastCommand == this.Command && (double)this.RepeatTimer.ElapsedTicks / (double)Stopwatch.Frequency < JvcProtocolHandler.RepeatCommandMaxLength) {
								this.OnCommandReceived(new InfraredCommandReceivedEventArgs(this.Command, ++this.LastCommandRepeatCount));
							} else {
								this.LastCommand = this.Command;
								this.OnCommandReceived(new InfraredCommandReceivedEventArgs(this.Command, this.LastCommandRepeatCount = 1));
							}
							this.RepeatTimer.Reset();
							this.RepeatTimer.Start();
						}
						this.Command = null;
					} else if (this.Command == null) {
						this.Command = new JvcCommand();
					}
				}
			} else {
				// We've just received a space.
				if (!this.ReceivingStartBit) {
					// This is a regular data bit.
					double PulseLength = this.BurstLength + time;
					if (PulseLength > JvcProtocolHandler.IntraBitMaxLength) {
						this.Command = null;
					}
					if (this.Command != null) {
						// Add the data bit.
						this.Command.AddMostSignificantBit(PulseLength > JvcProtocolHandler.DataBitLengthThreshold);
					}
				}
			}
		}

		#endregion

		#region Events
		
		public event EventHandler<InfraredCommandReceivedEventArgs> CommandReceived;

		protected virtual void OnCommandReceived(InfraredCommandReceivedEventArgs e) {
			if (this.CommandReceived != null) this.CommandReceived(this, e);
		}
		
		#endregion

	}
}