﻿using System;
using System.Diagnostics;

namespace BeeDevelopment.Infrared.Protocols {
	/// <summary>
	/// Decodes incoming old Panasonic commands.
	/// </summary>
	class OldPanasonicProtocolHandler : IRemoteControlProtocolHandler {

		#region Constants

		/// <summary>
		/// The minimum timestep length (nominally 420uS in the USA/Canada).
		/// </summary>
		private const double TimestepMinLength = 360E-6;

		/// <summary>
		/// The maximum timestep length (nominally 460uS in Europe).
		/// </summary>
		private const double TimestepMaxLength = 520E-6;

		/// <summary>
		/// The average length of a timestep.
		/// </summary>
		private const double TimestepAverageLength = (TimestepMinLength + TimestepMaxLength) / 2.0d;

		/// <summary>
		/// The minimum length of a start bit mark/space.
		/// </summary>
		private const double StartBitMinLength = TimestepMinLength * 8.0d;

		/// <summary>
		/// The maximum length of a start bit mark/space.
		/// </summary>
		private const double StartBitMaxLength = TimestepMaxLength * 8.0d;

		/// <summary>
		/// The minimum length for a "short" bit.
		/// </summary>
		private const double ShortBitMinLength = TimestepMinLength * 2.0d;

		/// <summary>
		/// The maximum length for a "short" bit.
		/// </summary>
		private const double ShortBitMaxLength = TimestepMaxLength * 2.0d;

		/// <summary>
		/// The minimum length for a "long" bit.
		/// </summary>
		private const double LongBitMinLength = TimestepMinLength * 6.0d;

		/// <summary>
		/// The maximum length for a "long" bit.
		/// </summary>
		private const double LongBitMaxLength = TimestepMaxLength * 6.0d;

		/// <summary>
		/// The threshold value between a "short" bit and a "long" bit.
		/// </summary>
		private const double LogicBitLengthThreshold = ((TimestepAverageLength * 2.0d) + (TimestepAverageLength * 6.0d)) / 2.0d;

		/// <summary>
		/// The maximum time length between repeating commands.
		/// </summary>
		private const double RepeatCommandMaxLength = 120.0E-3;

		#endregion

		#region Private Fields

		/// <summary>
		/// Set to <c>true</c> when receiving a start bit, <c>false</c> otherwise.
		/// </summary>
		private bool ReceivingStartBit = false;

		/// <summary>
		/// Stores the command as it gets built up.
		/// </summary>
		private OldPanasonicCommand Command = null;

		/// <summary>
		/// Stores the last received command.
		/// </summary>
		private OldPanasonicCommand LastCommand = null;

		/// <summary>
		/// Stores the number of times the received command has been repeated.
		/// </summary>
		private int LastCommandRepeatCount = 0;

		/// <summary>
		/// A <see cref="Stopwatch"/> instance used to time repeating commands.
		/// </summary>
		private Stopwatch RepeatTimer = null;

		#endregion

		public OldPanasonicProtocolHandler() {
			this.RepeatTimer = new Stopwatch();
			this.RepeatTimer.Start();
		}

		public void PinChanged(double time, bool newState) {
			if (newState) {
				// We've just received an IR burst (mark).
				if (time > OldPanasonicProtocolHandler.StartBitMinLength && time < OldPanasonicProtocolHandler.StartBitMaxLength) {
					this.ReceivingStartBit = true;
					this.Command = null;
				}
			} else {
				// We've just received a space.
				if (this.ReceivingStartBit) {
					if (time > OldPanasonicProtocolHandler.StartBitMinLength && time < OldPanasonicProtocolHandler.StartBitMaxLength) {
						this.Command = new OldPanasonicCommand();
					} else {
						this.Command = null;
					}
					this.ReceivingStartBit = false;
				} else {
					if (time < OldPanasonicProtocolHandler.ShortBitMinLength || time > OldPanasonicProtocolHandler.LongBitMaxLength) {
						this.Command = null;
					}
					if (this.Command != null) {
						// Is it a 0 bit or a 1 bit?
						this.Command.AddMostSignificantBit(time > OldPanasonicProtocolHandler.LogicBitLengthThreshold);
						// We've received a complete signal.
						if (this.Command.Count == 22) {
							if (this.Command.GetRange(0, 11) == ~this.Command.GetRange(11, 11)) {
								var ReceivedCommand = this.Command;
								if (ReceivedCommand != this.LastCommand || (double)this.RepeatTimer.ElapsedTicks / (double)Stopwatch.Frequency > OldPanasonicProtocolHandler.RepeatCommandMaxLength) {
									this.LastCommand = ReceivedCommand;
									this.LastCommandRepeatCount = 0;
								}
								this.OnCommandReceived(new InfraredCommandReceivedEventArgs(ReceivedCommand, ++this.LastCommandRepeatCount));
								this.RepeatTimer.Reset();
								this.RepeatTimer.Start();
							}
							this.Command = null;
						}
					}
				}
			}
		}

		public event EventHandler<InfraredCommandReceivedEventArgs> CommandReceived;

		protected virtual void OnCommandReceived(InfraredCommandReceivedEventArgs e) {
			if (this.CommandReceived != null) this.CommandReceived(this, e);
		}

	}
}
